# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: timeless/timeless.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..jobs import job_pb2 as jobs_dot_job__pb2
from ..jobs import dag_pb2 as jobs_dot_dag__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17timeless/timeless.proto\x12\x08timeless\x1a\x0ejobs/job.proto\x1a\x0ejobs/dag.proto\"F\n\x10RandomDAGRequest\x12\x11\n\tmax_depth\x18\x01 \x01(\x05\x12\x1f\n\x17max_children_per_parent\x18\x02 \x01(\x05\x32m\n\x08Timeless\x12\x39\n\x0cGetRandomDag\x12\x1a.timeless.RandomDAGRequest\x1a\r.jobs.dag.DAG\x12&\n\x0c\x46illJobsFlow\x12\t.jobs.Job\x1a\t.jobs.Job\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'timeless.timeless_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _RANDOMDAGREQUEST._serialized_start=69
  _RANDOMDAGREQUEST._serialized_end=139
  _TIMELESS._serialized_start=141
  _TIMELESS._serialized_end=250
# @@protoc_insertion_point(module_scope)



"""
This is CLI for `nbox`, it is meant to be as simple to use as possible. We using `python-fire` for building
our CLI, which means you can access underlying system with same commands as you use with python.

# SSH into Instances

You can SSH into your instance with the `nbx tunnel` command. This command will create a tunnel
to your instance and start an SSH session.

```bash
nbx tunnel 8000 -i "instance-id"
```
"""

import os
import fire
import requests
import webbrowser
from json import dumps
from shutil import rmtree
from typing import Dict, Any

import nbox.utils as U
from nbox.jobs import Job, Serve
from nbox.init import nbox_ws_v1
from nbox.auth import init_secret, AuthConfig, secret
from nbox.instance import Instance
from nbox.sub_utils.ssh import tunnel
from nbox.relics import Relics
from nbox.lmao import LmaoCLI
from nbox.version import __version__ as V
from nbox.nbxlib.fire import NBXFire
from nbox.projects import Project
from nbox.utils import logger, lo

class Config(object):
  def update(self, workspace_id: str):
    """Set global config for `nbox`"""
    secret = init_secret()
    data = secret(AuthConfig.cache)
    redo = not data or (workspace_id not in data)
    if redo:
      workspaces = requests.get(
        secret(AuthConfig.url) + f"/api/v1/workspace",
        headers = nbox_ws_v1._session.headers
      ).json()["data"]
      # workspaces = nbox_ws_v1.workspace()
      workspace_details = list(filter(lambda x: x["workspace_id"] == workspace_id, workspaces))
      if len(workspace_details) == 0:
        logger.error(f"Could not find the workspace ID: {workspace_id}. Please check the workspace ID and try again.")
        raise Exception("Invalid workspace ID")
      workspace_details = workspace_details[0]
      workspace_name = workspace_details["workspace_name"]
      secret.secrets[AuthConfig.cache].update({
        workspace_id: {
          "workspace_id": workspace_id,
          "workspace_name": workspace_name
        }
      })
      secret.put(AuthConfig.cache)
    else:
      data = data[workspace_id]
      workspace_name = data["workspace_name"]

    secret.put(AuthConfig.workspace_id, workspace_id, True)
    secret.put(AuthConfig.workspace_name, workspace_name, True)
    logger.info(f"Global Workspace: {workspace_name}")

  def show(self):
    """Pretty print global config for `nbox`"""
    logger.info(lo(
      "nbox config:",
      workspace_name = secret(AuthConfig.workspace_name),
      workspace_id = secret(AuthConfig.workspace_id),
      username = secret(AuthConfig.username),
      nbox = V,
      URL = secret(AuthConfig.url),
    ))

  def clear(self):
    cp = U.join(U.env.NBOX_HOME_DIR(), ".cache")
    logger.info(f"Clearing: {cp}")
    rmtree(cp)
    os.makedirs(cp)

def open_home():
  """Open current NBX platform"""
  webbrowser.open(secret(AuthConfig.url))


def get(api_end: str, no_pp: bool = False, **kwargs):
  """GET any API.

  Args:
    api_end (str): something like '/v1/api'
    no_pp (bool, optional): Pretty print. Defaults to False.
  """
  api_end = api_end.strip("/")
  if nbox_ws_v1 == None:
    raise RuntimeError("Not connected to NimbleBox.ai webserver")
  out = nbox_ws_v1
  for k in api_end.split("/"):
    out = getattr(out, k)
  res = out(_method = "get", **kwargs)
  if not no_pp:
    logger.info(dumps(res, indent=2))
  else:
    return res

def login(force: bool = False):
  """Re authenticate `nbox`. NOTE: Will remove all added keys to `secrets`"""
  if force:
    fp = U.join(U.env.NBOX_HOME_DIR(), "secrets.json")
    os.remove(fp)
    init_secret()
    return
  else:
    logger.info("Already logged in, to re-login use `nbx login --force`")

def version():
  logger.info(f"NimbleBox.ai Client Library\n    nbox version: {V}")


def why():
  print('''
Embrace the Future of Business with NimbleBox!

Artificial intelligence is the key to unlocking new opportunities in the 21st century.
With NimbleBox, you can simplify the building and management of AI operations and stay
ahead of the competition.

Join the AI revolution and seize the future!
''')


class NBXWS_CLI(object):
  def get(self, api: str, H: Dict[str, str] = {}):
    """make a GET call to any `api`, it can either contain '/' or '.'"""
    if nbox_ws_v1 == None:
      raise RuntimeError("Not connected to NimbleBox.ai webserver")
    api = api.strip("/")
    splitter = '.' if '.' in api else '/'
    nbox_ws_v1._session.headers.update(H)
    out = nbox_ws_v1
    for k in api.split(splitter):
      out = getattr(out, k)
    res = out(_method = "get")
    return res

  def post(self, api: str, d: Dict[str, Any] = {}, H: Dict[str, str] = {}):
    """make a POST call to any `api`, it can either contain '/' or '.'"""
    if nbox_ws_v1 == None:
      raise RuntimeError("Not connected to NimbleBox.ai webserver")
    api = api.strip("/")
    splitter = '.' if '.' in api else '/'
    nbox_ws_v1._session.headers.update(H)
    out = nbox_ws_v1
    for k in api.split(splitter):
      out = getattr(out, k)
    res = out(_method = "post", data = d)
    return res

  # def rapidoc(self):
  #   openapi = secret("openapi_spec", None)
  #   if openapi == None:
  #     raise RuntimeError("Not connected to NimbleBox.ai webserver")
  #   fp = U.join(U.folder(__file__), "assets", "rapidoc.html")
  #   with open(fp, "r") as f:
  #     temp = jinja2.Template(f.read())
  #   with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix=".html") as f:
  #     f.write(temp.render(openapi = dumps(openapi)))
  #     print(f.name)
  #     webbrowser.open(f"file://{f.name}")


def main():
  component = {
    "build"    : Instance,
    "config"   : Config,
    "get"      : get,
    "jobs"     : Job,
    "lmao"     : LmaoCLI,
    "login"    : login,
    "open"     : open_home,
    "projects" : Project,
    "relics"   : Relics,
    "serve"    : Serve,
    "tunnel"   : tunnel,
    "version"  : version,
    "why"      : why,
    "ws"       : NBXWS_CLI,
  }

  fire.Fire(component)
  # NBXFire(component)

if __name__ == "__main__":
  main()



"""
NimbleBox LMAO is our general purpose observability tool for any kind of computation you might have.
"""

from typing import Optional
from dateparser import parse as parse_date
from google.protobuf.timestamp_pb2 import Timestamp

from nbox.utils import logger
from nbox.auth import secret, AuthConfig
from nbox import messages as mpb

# all the sublime -> hyperloop stuff
from nbox.sublime.lmao_rpc_client import (
  ServingLogRequest, ServingLogResponse
)
from nbox.lmao import common


class LmaoCLI:

  def serving_logs(
    project_id: str,
    serving_id: str,
    key: str,
    after: Optional[str] = "",
    before: Optional[str] = "",
    limit: int = 100,
    f: str = "",
  ):
    """
    Get your logs from a serving by key and time range.

    ```bash
    nbx lmao serving logs '229dj92' '0283-nice-summer' field 'last 12 hours'
    nbx lmao serving logs '229dj92' '0283-nice-summer' field '2021-01-01' '2021-01-02'
    nbx lmao serving logs '229dj92' '0283-nice-summer' field '2021-01-01 12:00:00' '2021-01-02 12:00:00'
    nbx lmao serving logs '229dj92' '0283-nice-summer' field --limit 1000
    nbx lmao serving logs '229dj92' '0283-nice-summer' field --limit 1000 --f logs.json
    ```

    Args:
      project_id: ID of the project to which the serving belongs.
      serving_id: ID of the serving to which the log belongs.
      key: Key of the log to get.
      after: Return only logs after this timestamp.
      before: Return only logs before this timestamp.
      limit: Maximum number of logs to return.
      f: File to write logs to. If not provided, logs will be printed to stdout.
    """
    stub = common.get_lmao_stub()
    req = ServingLogRequest(
      workspace_id = secret(AuthConfig.workspace_id),
      project_id = project_id,
      serving_id = serving_id,
      key = key,
    )

    # date things
    if after:
      after_parsed = parse_date(after, settings={'TIMEZONE': 'UTC'})
      if after_parsed is None:
        raise ValueError(f"Invalid date format for `after` argument: {after}")
      t = Timestamp()
      t.FromDatetime(after_parsed)
      req.after.CopyFrom(t)
      logger.info(f"Will get logs from after: {after_parsed}")
    if before:
      before_parsed = parse_date(before, settings={'TIMEZONE': 'UTC'})
      if before_parsed is None:
        raise ValueError(f"Invalid date format for `before` argument: {before}")
      t = Timestamp()
      t.FromDatetime(before_parsed)
      req.before.CopyFrom(t)
      logger.info(f"Will get logs from before: {before_parsed}")
    if after and before:
      if after_parsed > before_parsed:
        raise ValueError(f"`after` date cannot be greater than `before` date")

    # limit things
    if limit < 0:
      raise ValueError("`limit` cannot be negative")
    req.limit = int(limit)

    logs: ServingLogResponse = stub.get_serving_log(req)
    logs_json = mpb.message_to_json(
      message = logs,
      including_default_value_fields = False,
      sort_keys = True,
      use_integers_for_enums = False
    )
    if not f:
      print(logs_json)
      return
    logger.info(f"Writing logs to {f}")
    with open(f, "w") as _f:
      _f.write(logs_json)

  # add all the CLIs at the bottom here:
  logs = {
    "serving": serving_logs,
  }



# Auto generated file. DO NOT EDIT.

import os
import requests
from functools import partial

from nbox.sublime.proto.relics_pb2 import *
from nbox.sublime.proto.common_pb2 import *
from nbox.sublime.proto.relics_rpc_pb2 import *

from nbox.sublime._yql.rest_pb2 import Echo
from nbox.sublime._yql.common import *

# ------ Stub ------ #

class RelicStore_Stub:
  def __init__(self, url: str, session: requests.Session = None):
    self.url = url.rstrip("/")
    self.session = session or requests.Session()
    self.status = partial(call_rpc, sess = self.session, url = f"{url}/")
    self.protos = partial(call_rpc, sess = self.session, url = f"{url}/protos")

  def create_relic(self, _CreateRelicRequest: CreateRelicRequest) -> Relic:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/create_relic",
      Echo(message = "CreateRelicRequest", base64_string=message_to_b64(_CreateRelicRequest), rpc_name = "create_relic")
    )
    if echo_resp is None:
      return None

    _Relic = Relic() # predefine the output proto
    _Relic = b64_to_message(echo_resp.base64_string, _Relic)
    return _Relic

  def list_relics(self, _ListRelicsRequest: ListRelicsRequest) -> ListRelicsResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/list_relics",
      Echo(message = "ListRelicsRequest", base64_string=message_to_b64(_ListRelicsRequest), rpc_name = "list_relics")
    )
    if echo_resp is None:
      return None

    _ListRelicsResponse = ListRelicsResponse() # predefine the output proto
    _ListRelicsResponse = b64_to_message(echo_resp.base64_string, _ListRelicsResponse)
    return _ListRelicsResponse

  def update_relic_meta(self, _Relic: Relic) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/update_relic_meta",
      Echo(message = "Relic", base64_string=message_to_b64(_Relic), rpc_name = "update_relic_meta")
    )
    if echo_resp is None:
      return None

    _Acknowledge = Acknowledge() # predefine the output proto
    _Acknowledge = b64_to_message(echo_resp.base64_string, _Acknowledge)
    return _Acknowledge

  def delete_relic(self, _Relic: Relic) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/delete_relic",
      Echo(message = "Relic", base64_string=message_to_b64(_Relic), rpc_name = "delete_relic")
    )
    if echo_resp is None:
      return None

    _Acknowledge = Acknowledge() # predefine the output proto
    _Acknowledge = b64_to_message(echo_resp.base64_string, _Acknowledge)
    return _Acknowledge

  def get_relic_details(self, _Relic: Relic) -> Relic:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_relic_details",
      Echo(message = "Relic", base64_string=message_to_b64(_Relic), rpc_name = "get_relic_details")
    )
    if echo_resp is None:
      return None

    _Relic = Relic() # predefine the output proto
    _Relic = b64_to_message(echo_resp.base64_string, _Relic)
    return _Relic

  def create_file(self, _RelicFile: RelicFile) -> RelicFile:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/create_file",
      Echo(message = "RelicFile", base64_string=message_to_b64(_RelicFile), rpc_name = "create_file")
    )
    if echo_resp is None:
      return None

    _RelicFile = RelicFile() # predefine the output proto
    _RelicFile = b64_to_message(echo_resp.base64_string, _RelicFile)
    return _RelicFile

  def list_relic_files(self, _ListRelicFilesRequest: ListRelicFilesRequest) -> ListRelicFilesResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/list_relic_files",
      Echo(message = "ListRelicFilesRequest", base64_string=message_to_b64(_ListRelicFilesRequest), rpc_name = "list_relic_files")
    )
    if echo_resp is None:
      return None

    _ListRelicFilesResponse = ListRelicFilesResponse() # predefine the output proto
    _ListRelicFilesResponse = b64_to_message(echo_resp.base64_string, _ListRelicFilesResponse)
    return _ListRelicFilesResponse

  def delete_relic_file(self, _RelicFile: RelicFile) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/delete_relic_file",
      Echo(message = "RelicFile", base64_string=message_to_b64(_RelicFile), rpc_name = "delete_relic_file")
    )
    if echo_resp is None:
      return None

    _Acknowledge = Acknowledge() # predefine the output proto
    _Acknowledge = b64_to_message(echo_resp.base64_string, _Acknowledge)
    return _Acknowledge

  def download_file(self, _RelicFile: RelicFile) -> RelicFile:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/download_file",
      Echo(message = "RelicFile", base64_string=message_to_b64(_RelicFile), rpc_name = "download_file")
    )
    if echo_resp is None:
      return None

    _RelicFile = RelicFile() # predefine the output proto
    _RelicFile = b64_to_message(echo_resp.base64_string, _RelicFile)
    return _RelicFile

  def get_activity_log(self, _ActivityLogRequest: ActivityLogRequest) -> ActivityLogResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_activity_log",
      Echo(message = "ActivityLogRequest", base64_string=message_to_b64(_ActivityLogRequest), rpc_name = "get_activity_log")
    )
    if echo_resp is None:
      return None

    _ActivityLogResponse = ActivityLogResponse() # predefine the output proto
    _ActivityLogResponse = b64_to_message(echo_resp.base64_string, _ActivityLogResponse)
    return _ActivityLogResponse


# ------ End Stub ------ #


import os
import sys
import ast
import shlex
import zipfile
from pprint import pformat
from subprocess import Popen

from nbox.utils import logger
from nbox import utils as U
from nbox.auth import secret, AuthConfig
from nbox.init import nbox_ws_v1
from nbox.lmao import (
  Lmao,
  LmaoLive,
  get_lmao_stub,
  get_git_details,
  lmao_v2_pb2 as pb,
  ExperimentConfig,
  LMAO_RM_PREFIX
)
from nbox.relics import Relics
from nbox.sublime.lmao_rpc_client import LMAO_Stub
from nbox.jobs import Job
from nbox.hyperloop.common.common_pb2 import Resource
from nbox.nbxlib.astea import Astea


class ProjectState:
  project_id: str = ""
  experiment_id: str = ""
  serving_id: str = ""


class Project:
  def __init__(self, id: str = ""):
    id = id or ProjectState.project_id
    logger.info(f"Connecting to Project: {id}")
    self.stub = nbox_ws_v1.projects.u(id)
    self.data = self.stub()
    self.pid = self.data["project_id"]
    self.lmao_stub = get_lmao_stub()
    self.project = self.lmao_stub.get_project(pb.Project(
      workspace_id = secret.workspace_id,
      project_id = self.pid
    ))
    self.relic = Relics(id = self.project.relic_id)
    self.workspace_id = secret.workspace_id

  def __repr__(self) -> str:
    return f"Project({self.pid})"

  def put_settings(self, project_name: str = "", project_description: str = ""):
    self.stub(
      "put",
      project_name = project_name or self.data["project_name"],
      project_description = project_description or self.data["project_description"]
    )
    self.data["project_name"] = project_name or self.data["project_name"]
    self.data["project_description"] = project_description or self.data["project_description"]

  def get(self, k):
    return self.data.get(k)

  def get_lmao_stub(self) -> LMAO_Stub:
    return self.lmao_stub

  def get_exp_tracker(self, experiment_id: str = "") -> Lmao:
    lmao = Lmao(project_id = self.pid, experiment_id = experiment_id or ProjectState.experiment_id)
    return lmao

  def get_live_tracker(self, serving_id: str = "") -> LmaoLive:
    lmao = LmaoLive(self.pid, serving_id = serving_id or ProjectState.serving_id)
    return lmao

  def get_relic(self) -> Relics:
    return self.relic

  def get_job_id(self) -> str:
    return self.data["job_list"][0]

  def get_deployment_id(self) -> str:
    return self.data["deployment_list"][0]

  # the big CLI command that replaces nbx lmao run

  def run(
    self,
    init_path: str,

    # all the arguments for the git files
    untracked: bool = False,
    untracked_no_limit: bool = False,

    # all the things for resources
    resource_cpu: str = "",
    resource_memory: str = "",
    resource_disk_size: str = "",
    resource_gpu: str = "",
    resource_gpu_count: str = "",
    resource_max_retries: int = 0,

    # any other things to pass to the function / class being called
    **run_kwargs,
  ):
    """Upload and register a new run for a NBX-LMAO project.

    Args:
      init_path (str): This can be a path to a `folder` or can be optionally of the structure `fp:fn` where `fp` is the path to the file and `fn` is the function name.
      untracked (bool, optional): If True, then untracked files below 1MB will also be zipped and uploaded. Defaults to False.
      untracked_no_limit (bool, optional): If True, then all untracked files will also be zipped and uploaded. Defaults to False.
      resource_cpu (str, optional): The CPU resource to allocate to the run. It can be found on your NimbleBox dashboard.
      resource_memory (str, optional): The memory resource to allocate to the run. It can be found on your NimbleBox dashboard.
      resource_disk_size (str, optional): The disk size resource to allocate to the run. It can be found on your NimbleBox dashboard.
      resource_gpu (str, optional): The GPU resource to use. It can be found on your NimbleBox dashboard.
      resource_gpu_count (str, optional): Number of GPUs allocated to the experiment. It can be found on your NimbleBox dashboard.
      resource_max_retries (int, optional): The maximum number of retries for a run. It can be found on your NimbleBox dashboard.
      **run_kwargs: These are the kwargs that will be passed to your function.
    """
    workspace_id = secret(AuthConfig.workspace_id)

    # fix data type conversion caused by the CLI
    resource_cpu = str(resource_cpu)
    resource_memory = str(resource_memory)
    resource_disk_size = str(resource_disk_size)
    resource_gpu = str(resource_gpu)
    resource_gpu_count = str(resource_gpu_count)
    resource_max_retries = int(resource_max_retries)

    # reconstruct the entire CLI command so we can show it in the UI
    reconstructed_cli_comm = (
      f"nbx projects --id '{self.pid}' - run '{init_path}'" +
      (f" --resource_cpu '{resource_cpu}'" if resource_cpu else "") +
      (f" --resource_memory '{resource_memory}'" if resource_memory else "") +
      (f" --resource_disk_size '{resource_disk_size}'" if resource_disk_size else "") +
      (f" --resource_gpu '{resource_gpu}'" if resource_gpu else "") +
      (f" --resource_gpu_count '{resource_gpu_count}'" if resource_gpu_count else "") +
      (f" --resource_max_retries {resource_max_retries}" if resource_max_retries else "")
    )
    if untracked_no_limit and not untracked:
      raise ValueError("Cannot set untracked_no_limit to True if untracked is False")
    if untracked:
      reconstructed_cli_comm += " --untracked"
    if untracked_no_limit:
      reconstructed_cli_comm += " --untracked_no_limit"
    for k, v in run_kwargs.items():
      if type(v) == bool:
        reconstructed_cli_comm += f" --{k}"
      else:
        reconstructed_cli_comm += f" --{k} '{v}'"
    logger.debug(f"command: {reconstructed_cli_comm}")

    # get the job
    job = Job(job_id = self.get_job_id())
    job.job_proto.resource.MergeFrom(Resource(
      cpu = resource_cpu,
      memory = resource_memory,
      disk_size = resource_disk_size,
      gpu = resource_gpu,
      gpu_count = resource_gpu_count,
      max_retries = resource_max_retries,
    ))

    # analyse the input file and function, extract the types and build the run_kwargs
    fn_file, fn_name = init_path.split(":")
    if not os.path.exists(fn_file + ".py"):
      raise Exception(f"File {fn_file} does not exist")
    init_folder, fn_file = os.path.split(fn_file + ".py")
    init_folder = init_folder or "."
    tea = Astea(fn_file)
    items = tea.find(fn_name)
    if not len(items):
      raise Exception(f"Function {fn_name} not found in {fn_file}")
    fn = items[0]
    defaults = fn.node.args.defaults
    args = fn.node.args.args
    _default_kwarg = {}
    for k,v in zip(args[::-1], defaults[::-1]):
      _k = k.arg
      if not type(v) == ast.Constant:
        raise Exception(f"Default value for {_k} is not a primitive type [int, float, str, bool] but {type(v)}")
      _default_kwarg[_k] = v.value
    args_set = set([x.arg for x in args])
    args_na = args_set - set(_default_kwarg.keys())
    args_not_passed = set(args_na) - set(run_kwargs.keys())
    if len(args_not_passed):
      raise Exception(f"Following arguments are not passed: {args_not_passed} but required, this can cause errors during execution")
    extra_args = set(run_kwargs.keys()) - args_set
    if len(extra_args) and not fn.node.args.kwarg:
      raise Exception(f"Following arguments are passed but not consumed by {fn_name}: {extra_args}")
    final_dict = {}
    for k,v in _default_kwarg.items():
      final_dict[k] = type(v)(run_kwargs.get(k, v))
    run_kwargs = final_dict

    # initialise the run with the big object and get the experiment_id
    if os.path.exists(U.join(init_folder, ".git")):
      git_det = get_git_details(init_folder)
    else:
      git_det = {}
    run = self.lmao_stub.init_run(pb.InitRunRequest(
      workspace_id = workspace_id,
      project_id = self.pid,
      agent_details = pb.AgentDetails(
        type = pb.AgentDetails.NBX.JOB,
        nbx_job_id = job.id, # run id will have to be updated from the job
      ),
      config = ExperimentConfig(
        run_kwargs = run_kwargs,
        git = git_det,
        resource = job.job_proto.resource,
        cli_comm = reconstructed_cli_comm,
        save_to_relic = True,
        enable_system_monitoring = False,
      ).to_json(),
    ))
    logger.info(f"Run ID: {run.experiment_id}")

    # create a git patch and upload it to relics
    if git_det:
      _zf = U.join(init_folder, "untracked.zip")
      zip_file = zipfile.ZipFile(_zf, "w", zipfile.ZIP_DEFLATED)
      patch_file = U.join(init_folder, "nbx_auto_patch.diff")
      f = open(patch_file, "w")
      Popen(shlex.split(f"git diff {' '.join(git_det['uncommited_files'])}"), stdout=f, stderr=sys.stderr).wait()
      f.close()
      zip_file.write(patch_file, arcname = patch_file)
      if untracked:
        untracked_files = git_det["untracked_files"] # what to do with these?
        if untracked_files:
          # see if any file is larger than 10MB and if so, warn the user
          warn_once = False
          for f in untracked_files:
            if os.path.getsize(f) > 1e7 and not untracked_no_limit:
              logger.warning(f"File: {f} is larger than 10MB and will not be available in sync")
              logger.warning("  Fix: use git to track small files, avoid large files")
              logger.warning("  Fix: nbox.Relics can be used to store large files")
              logger.warning("  Fix: use --untracked-no-limit to upload all files")
              warn_once = True
              continue
            zip_file.write(f, arcname = f)

      # upload the patch file to relics
      self.relic.put_to(_zf, f"{run.experiment_id}/git.zip")
      os.remove(_zf)
      os.remove(patch_file)

    # tell the server that this run is being scheduled so atleast the information is visible on the dashboard
    Job.upload(
      init_folder = init_path,
      id = job.id,

      # pass along the resource requirements
      resource_cpu = resource_cpu,
      resource_memory = resource_memory,
      resource_disk_size = resource_disk_size,
      resource_gpu = resource_gpu,
      resource_gpu_count = resource_gpu_count,
      resource_max_retries = resource_max_retries,
    )

    # now trigger the experiment
    fp = f"{run.project_id}/{run.experiment_id}"
    tag = f"{LMAO_RM_PREFIX}{fp}"
    logger.debug(f"Running job '{job.name}' ({job.id}) with tag: {tag}")
    job.trigger(tag)

    # finally print the location of the run where the users can track this
    logger.info(f"Run location: {secret(AuthConfig.url)}/workspace/{run.workspace_id}/projects/{run.project_id}#Experiments")



# Auto generated file. DO NOT EDIT.

import os
import requests
from functools import partial

from google.protobuf.timestamp_pb2 import *
from nbox.sublime.proto.lmao_v2_pb2 import *

from nbox.sublime._yql.rest_pb2 import Echo
from nbox.sublime._yql.common import *

# ------ Stub ------ #

class LMAO_Stub:
  def __init__(self, url: str, session: requests.Session = None):
    self.url = url.rstrip("/")
    self.session = session or requests.Session()
    self.status = partial(call_rpc, sess = self.session, url = f"{url}/")
    self.protos = partial(call_rpc, sess = self.session, url = f"{url}/protos")

  def init_project(self, _lmao_v2_InitProjectRequest: InitProjectRequest) -> InitProjectResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/init_project",
      Echo(message = "InitProjectRequest", base64_string=message_to_b64(_lmao_v2_InitProjectRequest), rpc_name = "init_project")
    )
    if echo_resp is None:
      return None

    _lmao_v2_InitProjectResponse = InitProjectResponse() # predefine the output proto
    _lmao_v2_InitProjectResponse = b64_to_message(echo_resp.base64_string, _lmao_v2_InitProjectResponse)
    return _lmao_v2_InitProjectResponse

  def get_project(self, _lmao_v2_Project: Project) -> Project:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_project",
      Echo(message = "Project", base64_string=message_to_b64(_lmao_v2_Project), rpc_name = "get_project")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Project = Project() # predefine the output proto
    _lmao_v2_Project = b64_to_message(echo_resp.base64_string, _lmao_v2_Project)
    return _lmao_v2_Project

  def list_projects(self, _lmao_v2_ListProjectsRequest: ListProjectsRequest) -> ListProjectsResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/list_projects",
      Echo(message = "ListProjectsRequest", base64_string=message_to_b64(_lmao_v2_ListProjectsRequest), rpc_name = "list_projects")
    )
    if echo_resp is None:
      return None

    _lmao_v2_ListProjectsResponse = ListProjectsResponse() # predefine the output proto
    _lmao_v2_ListProjectsResponse = b64_to_message(echo_resp.base64_string, _lmao_v2_ListProjectsResponse)
    return _lmao_v2_ListProjectsResponse

  def delete_project(self, _lmao_v2_Project: Project) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/delete_project",
      Echo(message = "Project", base64_string=message_to_b64(_lmao_v2_Project), rpc_name = "delete_project")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def init_run(self, _lmao_v2_InitRunRequest: InitRunRequest) -> Run:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/init_run",
      Echo(message = "InitRunRequest", base64_string=message_to_b64(_lmao_v2_InitRunRequest), rpc_name = "init_run")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Run = Run() # predefine the output proto
    _lmao_v2_Run = b64_to_message(echo_resp.base64_string, _lmao_v2_Run)
    return _lmao_v2_Run

  def update_run_status(self, _lmao_v2_Run: Run) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/update_run_status",
      Echo(message = "Run", base64_string=message_to_b64(_lmao_v2_Run), rpc_name = "update_run_status")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def on_log(self, _lmao_v2_RunLog: RunLog) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/on_log",
      Echo(message = "RunLog", base64_string=message_to_b64(_lmao_v2_RunLog), rpc_name = "on_log")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def on_train_end(self, _lmao_v2_Run: Run) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/on_train_end",
      Echo(message = "Run", base64_string=message_to_b64(_lmao_v2_Run), rpc_name = "on_train_end")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def list_runs(self, _lmao_v2_ListRunsRequest: ListRunsRequest) -> ListRunsResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/list_runs",
      Echo(message = "ListRunsRequest", base64_string=message_to_b64(_lmao_v2_ListRunsRequest), rpc_name = "list_runs")
    )
    if echo_resp is None:
      return None

    _lmao_v2_ListRunsResponse = ListRunsResponse() # predefine the output proto
    _lmao_v2_ListRunsResponse = b64_to_message(echo_resp.base64_string, _lmao_v2_ListRunsResponse)
    return _lmao_v2_ListRunsResponse

  def get_run_details(self, _lmao_v2_Run: Run) -> Run:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_run_details",
      Echo(message = "Run", base64_string=message_to_b64(_lmao_v2_Run), rpc_name = "get_run_details")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Run = Run() # predefine the output proto
    _lmao_v2_Run = b64_to_message(echo_resp.base64_string, _lmao_v2_Run)
    return _lmao_v2_Run

  def get_run_log(self, _lmao_v2_RunLogRequest: RunLogRequest) -> RunLog:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_run_log",
      Echo(message = "RunLogRequest", base64_string=message_to_b64(_lmao_v2_RunLogRequest), rpc_name = "get_run_log")
    )
    if echo_resp is None:
      return None

    _lmao_v2_RunLog = RunLog() # predefine the output proto
    _lmao_v2_RunLog = b64_to_message(echo_resp.base64_string, _lmao_v2_RunLog)
    return _lmao_v2_RunLog

  def get_experiment_table(self, _lmao_v2_GetExperimentTableRequest: GetExperimentTableRequest) -> ExperimentTable:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_experiment_table",
      Echo(message = "GetExperimentTableRequest", base64_string=message_to_b64(_lmao_v2_GetExperimentTableRequest), rpc_name = "get_experiment_table")
    )
    if echo_resp is None:
      return None

    _lmao_v2_ExperimentTable = ExperimentTable() # predefine the output proto
    _lmao_v2_ExperimentTable = b64_to_message(echo_resp.base64_string, _lmao_v2_ExperimentTable)
    return _lmao_v2_ExperimentTable

  def delete_experiment(self, _lmao_v2_Run: Run) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/delete_experiment",
      Echo(message = "Run", base64_string=message_to_b64(_lmao_v2_Run), rpc_name = "delete_experiment")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def init_serving(self, _lmao_v2_InitRunRequest: InitRunRequest) -> Serving:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/init_serving",
      Echo(message = "InitRunRequest", base64_string=message_to_b64(_lmao_v2_InitRunRequest), rpc_name = "init_serving")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Serving = Serving() # predefine the output proto
    _lmao_v2_Serving = b64_to_message(echo_resp.base64_string, _lmao_v2_Serving)
    return _lmao_v2_Serving

  def update_serving_status(self, _lmao_v2_Serving: Serving) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/update_serving_status",
      Echo(message = "Serving", base64_string=message_to_b64(_lmao_v2_Serving), rpc_name = "update_serving_status")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def on_serving_log(self, _lmao_v2_RunLog: RunLog) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/on_serving_log",
      Echo(message = "RunLog", base64_string=message_to_b64(_lmao_v2_RunLog), rpc_name = "on_serving_log")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def on_serving_end(self, _lmao_v2_Serving: Serving) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/on_serving_end",
      Echo(message = "Serving", base64_string=message_to_b64(_lmao_v2_Serving), rpc_name = "on_serving_end")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def list_servings(self, _lmao_v2_ListServingsRequest: ListServingsRequest) -> ListServingsResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/list_servings",
      Echo(message = "ListServingsRequest", base64_string=message_to_b64(_lmao_v2_ListServingsRequest), rpc_name = "list_servings")
    )
    if echo_resp is None:
      return None

    _lmao_v2_ListServingsResponse = ListServingsResponse() # predefine the output proto
    _lmao_v2_ListServingsResponse = b64_to_message(echo_resp.base64_string, _lmao_v2_ListServingsResponse)
    return _lmao_v2_ListServingsResponse

  def get_serving_details(self, _lmao_v2_Serving: Serving) -> Serving:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_serving_details",
      Echo(message = "Serving", base64_string=message_to_b64(_lmao_v2_Serving), rpc_name = "get_serving_details")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Serving = Serving() # predefine the output proto
    _lmao_v2_Serving = b64_to_message(echo_resp.base64_string, _lmao_v2_Serving)
    return _lmao_v2_Serving

  def get_serving_log(self, _lmao_v2_ServingLogRequest: ServingLogRequest) -> ServingLogResponse:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_serving_log",
      Echo(message = "ServingLogRequest", base64_string=message_to_b64(_lmao_v2_ServingLogRequest), rpc_name = "get_serving_log")
    )
    if echo_resp is None:
      return None

    _lmao_v2_ServingLogResponse = ServingLogResponse() # predefine the output proto
    _lmao_v2_ServingLogResponse = b64_to_message(echo_resp.base64_string, _lmao_v2_ServingLogResponse)
    return _lmao_v2_ServingLogResponse

  def delete_serving(self, _lmao_v2_Serving: Serving) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/delete_serving",
      Echo(message = "Serving", base64_string=message_to_b64(_lmao_v2_Serving), rpc_name = "delete_serving")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def get_rule_builder(self, _lmao_v2_RuleBuilder: RuleBuilder) -> RuleBuilder:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/get_rule_builder",
      Echo(message = "RuleBuilder", base64_string=message_to_b64(_lmao_v2_RuleBuilder), rpc_name = "get_rule_builder")
    )
    if echo_resp is None:
      return None

    _lmao_v2_RuleBuilder = RuleBuilder() # predefine the output proto
    _lmao_v2_RuleBuilder = b64_to_message(echo_resp.base64_string, _lmao_v2_RuleBuilder)
    return _lmao_v2_RuleBuilder

  def create_rule(self, _lmao_v2_Rule: Rule) -> Rule:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/create_rule",
      Echo(message = "Rule", base64_string=message_to_b64(_lmao_v2_Rule), rpc_name = "create_rule")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Rule = Rule() # predefine the output proto
    _lmao_v2_Rule = b64_to_message(echo_resp.base64_string, _lmao_v2_Rule)
    return _lmao_v2_Rule

  def list_rules(self, _lmao_v2_ListRules: ListRules) -> ListRules:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/list_rules",
      Echo(message = "ListRules", base64_string=message_to_b64(_lmao_v2_ListRules), rpc_name = "list_rules")
    )
    if echo_resp is None:
      return None

    _lmao_v2_ListRules = ListRules() # predefine the output proto
    _lmao_v2_ListRules = b64_to_message(echo_resp.base64_string, _lmao_v2_ListRules)
    return _lmao_v2_ListRules

  def delete_rule(self, _lmao_v2_Rule: Rule) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/delete_rule",
      Echo(message = "Rule", base64_string=message_to_b64(_lmao_v2_Rule), rpc_name = "delete_rule")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge

  def update_rule(self, _lmao_v2_Rule: Rule) -> Acknowledge:
    echo_resp: Echo = call_rpc(
      self.session,
      f"{self.url}/update_rule",
      Echo(message = "Rule", base64_string=message_to_b64(_lmao_v2_Rule), rpc_name = "update_rule")
    )
    if echo_resp is None:
      return None

    _lmao_v2_Acknowledge = Acknowledge() # predefine the output proto
    _lmao_v2_Acknowledge = b64_to_message(echo_resp.base64_string, _lmao_v2_Acknowledge)
    return _lmao_v2_Acknowledge


# ------ End Stub ------ #